/* data-form.c
 *
 * Copyright (C) 2002 Vivien Malerba
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include "data-form.h"
#include "marshal.h"
#include "data-entry.h"

static void data_form_class_init (DataFormClass * class);
static void data_form_init (DataForm * form);
static void data_form_finalize (GObject   * object);

enum
{
	ACTION_REFRESH,
	ACTION_INSERT,
	ACTION_VIEWALL,
	ACTION_CHOOSE,
	CONTENTS_MODIFIED,
	LAST_SIGNAL
};


typedef enum   _DataFormMode   DataFormMode;
typedef struct _ValueRef       ValueRef;
typedef struct _EntryEntity    EntryEntity;

enum _DataFormMode {
	/* edition of data from a query, for a given DbTable (for which there is only one 
	   corresponding QueryView object. The only data entry widgets are the ones related
	   to the DbTable */
        DATA_FORM_EDIT_MODE,

	/* same as EDIT_MODE but all entries are empty (or imposed by the Query) to start with */
        DATA_FORM_INSERT_MODE,

	/* data selection mode: selects data from a given Query and an given QueryView in this
	   Query: displays choice combo entries. It creates its own Query and also managed its own
	   ServerResultset object. */
        DATA_FORM_SELECT_MODE, 

	/* just showing the data from the ServerResultset of the QueryExec object, no modification
	   allowed (if modifications are required, then create another DataForm in EDIT or INSERT 
	   mode*/
        DATA_FORM_BROWSE_MODE  /* same as EDIT but no modification allowed */
};

struct _DataFormPriv
{
	DataFormMode     mode;

	QueryExec       *qx;       /* not NULL */
	QueryView       *qv;

        GSList          *value_refs;
        GSList          *entries;

	GSList          *children;      /* sub forms */

	/* SELECT mode only */
	DataForm        *master;
	Query           *internal_query;
	ServerResultset *internal_rs;

	/* EDIT and BROWSE modes only */
        gulong           numrow;
	GtkWidget       *top_notebook;

	/* EDIT, INSERT and BROWSE modes only */
	GtkWidget       *actions_area;
	guint            actions; /* possible user actions */
};

struct _ValueRef {
        gpointer        ref;       /* DbField or QueryParamValue */
        EntryEntity    *entry;
};
#define VALUE_REF(vr) ((ValueRef*)(vr))


struct _EntryEntity {
        ValueRef       *value_ref;
        QueryField     *qf;

        guint           numcol; /* if SELECT, refers to the internal_rs and not the RS of QueryExec */

	/* if DataEntry only */
        DataEntry      *entry;

	/* if other DataForm only */
	DataForm       *form;
	QueryField     *join_to;
};
#define ENTRY_ENTITY(ee) ((EntryEntity*)(ee))


/* Returns the value of a QueryField from a Form:
   the QF needs to be a field and reference a QueryView which is represented 
   by the DataForm df.
   returns a new GdaValue, or NULL if there is a problem (along with a warning)
*/
static GdaValue *data_form_get_gdavalue (DataForm *df, QueryField *qf);

/* Tries to find a DataForm for a given QueryView. */
static DataForm *data_form_find_from_query_view (DataForm *master, QueryView *qv);

static ValueRef *data_form_find_value_ref (DataForm *df, gpointer ref);

/* Builds the default builtin Form from the internal structure of the DataForm */
static GtkWidget *build_complete_noglade_form_widget (DataForm * form);

static void     data_form_user_action_cb (GtkWidget *wid, DataForm * form);

static gint data_form_signals[LAST_SIGNAL] = { 0, 0, 0, 0, 0 };

/* get a pointer to the parents to be able to call their destructor */
static GObjectClass *parent_class = NULL;

guint
data_form_get_type (void)
{
	static GType type = 0;

	if (!type) {
		static const GTypeInfo info = {
			sizeof (DataFormClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) data_form_class_init,
			NULL,
			NULL,
			sizeof (DataForm),
			0,
			(GInstanceInitFunc) data_form_init
		};		

		type = g_type_register_static (GTK_TYPE_VBOX, "DataForm", &info, 0);
	}

	return type;
}

static void
data_form_class_init (DataFormClass * class)
{
	GObjectClass   *object_class = G_OBJECT_CLASS (class);
	
	parent_class = g_type_class_peek_parent (class);

	data_form_signals[ACTION_REFRESH] =
		g_signal_new ("action_refresh",
			      G_TYPE_FROM_CLASS (object_class),
			      G_SIGNAL_RUN_FIRST,
			      G_STRUCT_OFFSET (DataFormClass, action_refresh),
			      NULL, NULL,
			      marshal_VOID__VOID, G_TYPE_NONE,
			      0);
	data_form_signals[ACTION_INSERT] =
		g_signal_new ("action_insert",
			      G_TYPE_FROM_CLASS (object_class),
			      G_SIGNAL_RUN_FIRST,
			      G_STRUCT_OFFSET (DataFormClass, action_insert),
			      NULL, NULL,
			      marshal_VOID__VOID, G_TYPE_NONE,
			      0);
	data_form_signals[ACTION_VIEWALL] =
		g_signal_new ("action_viewall",
			      G_TYPE_FROM_CLASS (object_class),
			      G_SIGNAL_RUN_FIRST,
			      G_STRUCT_OFFSET (DataFormClass, action_viewall),
			      NULL, NULL,
			      marshal_VOID__VOID, G_TYPE_NONE,
			      0);
	data_form_signals[ACTION_CHOOSE] =
		g_signal_new ("action_choose",
			      G_TYPE_FROM_CLASS (object_class),
			      G_SIGNAL_RUN_FIRST,
			      G_STRUCT_OFFSET (DataFormClass, action_choose),
			      NULL, NULL,
			      marshal_VOID__ULONG, G_TYPE_NONE, 1,
			      G_TYPE_ULONG);
	data_form_signals[CONTENTS_MODIFIED] =
		g_signal_new ("contents_modified",
			      G_TYPE_FROM_CLASS (object_class),
			      G_SIGNAL_RUN_FIRST,
			      G_STRUCT_OFFSET (DataFormClass, contents_modified),
			      NULL, NULL,
			      marshal_VOID__VOID, G_TYPE_NONE,
			      0);

	class->action_refresh = NULL;
	class->action_insert = NULL;
	class->action_viewall = NULL;
	class->action_choose = NULL;
	class->contents_modified = NULL;
	object_class->finalize = data_form_finalize;
}

static void
data_form_init (DataForm * form)
{
	form->priv = g_new0 (DataFormPriv, 1);
	form->priv->qx = NULL;
	form->priv->qv = NULL;
	form->priv->value_refs = NULL;
	form->priv->entries = NULL;
	form->priv->master = NULL;
	form->priv->children = NULL;
	form->priv->top_notebook = NULL;
	form->priv->internal_query = NULL;
	form->priv->internal_rs = NULL;
	form->priv->numrow = 0;
	form->priv->actions = 0;
	form->priv->actions_area = NULL;
}


static void data_form_edit_initialize (DataForm * form);
GtkWidget *
data_form_new_edit (QueryExec *qx, gulong numrow)
{
	GObject   *obj;
	DataForm *form;
	GSList *list;
	gint i;

	g_return_val_if_fail (qx && IS_QUERY_EXEC (qx), NULL);

	obj = g_object_new (DATA_FORM_TYPE, NULL);
	form = DATA_FORM (obj);

	form->priv->qx = qx;
	form->priv->mode = DATA_FORM_EDIT_MODE;
	form->priv->numrow = numrow;
	form->priv->master = form; /* this form is the master for the other sub forms */
	form->priv->actions = query_exec_get_user_actions (qx);

	/* Find the right QueryView */
	list = query_exec_get_query_ref(qx)->views;
	i = 0;
	while (list) {
		QueryView *qv = QUERY_VIEW (list->data);
		if (IS_DB_TABLE (query_view_get_view_obj (qv)) && 
		    (DB_TABLE (query_view_get_view_obj (qv)) == query_exec_get_modif_table (qx))) {
			form->priv->qv = qv;
			i++;
		}
		list = g_slist_next (list);
	}
	if (i>1)
		g_warning ("There is more than 1 QueryView for the edited table!");

	data_form_edit_initialize (form);

	return GTK_WIDGET (obj);
}

static void data_form_insert_initialize (DataForm * form);
GtkWidget *
data_form_new_insert (QueryExec *qx)
{
	GObject   *obj;
	DataForm *form;
	GSList *list;
	gint i;

	g_return_val_if_fail (qx && IS_QUERY_EXEC (qx), NULL);

	obj = g_object_new (DATA_FORM_TYPE, NULL);
	form = DATA_FORM (obj);

	form->priv->qx = qx;
	form->priv->mode = DATA_FORM_INSERT_MODE;
	form->priv->master = form; /* this form is the master for the other sub forms */
	form->priv->actions = QUERY_ACTION_COMMIT;

	/* Find the right QueryView */
	list = query_exec_get_query_ref(qx)->views;
	i = 0;
	while (list) {
		QueryView *qv = QUERY_VIEW (list->data);
		if (IS_DB_TABLE (query_view_get_view_obj (qv)) && 
		    (DB_TABLE (query_view_get_view_obj (qv)) == query_exec_get_modif_table (qx))) {
			form->priv->qv = qv;
			i++;
		}
		list = g_slist_next (list);
	}
	if (i>1)
		g_warning ("There is more than 1 QueryView for the edited table!");

	data_form_insert_initialize (form);

	return GTK_WIDGET (obj);
}


static void data_form_browse_initialize (DataForm * form);
GtkWidget *
data_form_new_browse (QueryExec *qx, gulong numrow)
{
	GObject   *obj;
	DataForm *form;

	g_return_val_if_fail (qx && IS_QUERY_EXEC (qx), NULL);

	/* FIXME: how to set the parameters for the GtkVBox? */
	obj = g_object_new (DATA_FORM_TYPE, NULL);
	form = DATA_FORM (obj);

	form->priv->qx = qx;
	form->priv->mode = DATA_FORM_BROWSE_MODE;
	form->priv->numrow = numrow;

	data_form_browse_initialize (form);

	return GTK_WIDGET (obj);
}


static void data_form_select_initialize (DataForm * form);
GtkWidget *
data_form_new_select (QueryExec *qx, QueryView *qv, gulong numrow)
{
	GObject   *obj;
	DataForm *form;

	g_return_val_if_fail (qx && IS_QUERY_EXEC (qx), NULL);
	g_return_val_if_fail (qv && IS_QUERY_VIEW (qv), NULL);
	g_return_val_if_fail (query_view_get_query (QUERY_VIEW (qv)) != query_exec_get_query_ref (qx), NULL);

	/* FIXME: how to set the parameters for the GtkVBox? */
	obj = g_object_new (DATA_FORM_TYPE, NULL);
	form = DATA_FORM (obj);

	form->priv->qx = qx;
	form->priv->mode = DATA_FORM_SELECT_MODE;
	form->priv->numrow = numrow;

	data_form_select_initialize (form);

	return GTK_WIDGET (obj);
}

static void
data_form_finalize (GObject   * object)
{
	DataForm *df;
	GSList *list;

	g_return_if_fail (object && IS_DATA_FORM (object));

	df = DATA_FORM (object);

	if (df->priv) {
		/* ValueRef */
		if (df->priv->value_refs) {
			list = df->priv->value_refs;
			while (list) {
				g_free (list->data);
				list = g_slist_next (list);
			}
			g_slist_free (df->priv->value_refs);
			df->priv->value_refs = NULL;
		}
		
		/* Entries */
		if (df->priv->entries) {
			list = df->priv->entries;
			while (list) {
				g_free (list->data);
				list = g_slist_next (list);
			}
			g_slist_free (df->priv->entries);
			df->priv->entries = NULL;
		}
		
		/* Children */
		if (df->priv->children) {
			g_slist_free (df->priv->children);
			df->priv->children = NULL;
		}
		
		/* Internal query */
		if (df->priv->internal_query) {
			g_object_unref (G_OBJECT (df->priv->internal_query));
			df->priv->internal_query = NULL;
		}
		
		/* Internal ServerResultset */
		if (df->priv->internal_rs) {
			g_object_unref (G_OBJECT (df->priv->internal_rs));
			df->priv->internal_rs = NULL;
		}
		g_free (df->priv);
		df->priv = NULL;
	}

	/* for the parent class */
	parent_class->finalize (object);
}












/*
 *
 * EDIT MODE
 *
 */
static GtkWidget *data_form_add_action_buttons (DataForm *form, GCallback cb, gpointer data);
static void build_edit_insert_form_structure (DataForm * form);
static void 
data_form_edit_initialize (DataForm * form)
{
	GtkWidget *topwid, *wid, *buttons;

	/* notebook to show either "No Data" or a form if there are some data */
	topwid = gtk_notebook_new ();
	gtk_notebook_set_show_tabs (GTK_NOTEBOOK (topwid), FALSE);

	/* First page, "No Data" */
	wid = gtk_label_new (_("No Data!"));
	gtk_notebook_append_page (GTK_NOTEBOOK (topwid), wid, NULL);
	gtk_widget_show (wid);


	/* Second Page, the form itself */
	build_edit_insert_form_structure (form);
	if (0) {
		/* FIXME: update when glade forms are introduced */
		/* we have a custom glade form here */
		wid = gtk_label_new ("Glade imported form to create");
		gtk_widget_show (wid);
	}
	else {
		/* default builtin form */
		wid = build_complete_noglade_form_widget (form);
		gtk_widget_show (wid);
	}
	gtk_notebook_append_page (GTK_NOTEBOOK (topwid), wid, NULL);
	form->priv->top_notebook = topwid;


	/* Action buttons, etc */
	gtk_widget_show (topwid);
	gtk_box_pack_start_defaults (GTK_BOX (form), topwid);
	
	buttons = data_form_add_action_buttons (form, G_CALLBACK (data_form_user_action_cb), form);
	gtk_widget_show_all (buttons);
	gtk_box_pack_start (GTK_BOX (form), buttons, FALSE, FALSE, GNOME_PAD/2.);


	/* Default shown page */
	if (! query_exec_get_data_set (form->priv->qx) || 
	    (server_resultset_get_nbtuples (query_exec_get_data_set (form->priv->qx)) == 0))
		gtk_notebook_set_current_page (GTK_NOTEBOOK (topwid), 0);
	else 
		gtk_notebook_set_current_page (GTK_NOTEBOOK (topwid), 1);
}

static void 
build_edit_insert_form_structure (DataForm * form)
{
	QueryField *qf;
	GSList *list;
	
	list = query_exec_get_query_ref (form->priv->qx)->fields;
	while (list) {
		qf = QUERY_FIELD (list->data);

		if (query_field_get_ftype (qf) == QUERY_FIELD_FIELD) {
			QueryView *qv;

			qv = query_field_field_get_queryview (qf);
			if (qv && IS_DB_TABLE (query_view_get_view_obj (qv)) && 
			    (DB_TABLE (query_view_get_view_obj (qv)) == query_exec_get_modif_table (form->priv->qx))) {
				GSList *joins;
				QueryJoin *qj = NULL;
				
				/* Find a join if it exists */
				joins = query_find_equi_join_with_end (query_exec_get_query_ref (form->priv->qx), qf);
				if (joins) {
					if (g_slist_length (joins) > 1)
						g_warning ("More than one Foreign Key on the %s field, ignored",
							   query_field_field_get_db_field (qf)->name);
					else
						qj = QUERY_JOIN (joins->data);
					g_slist_free (joins);
				}
				
				if (qj) { /* We create a sub form for the QueryJoin */
					QueryView *oqv;
					QueryField *oqf = NULL;
					DbField *ofield = NULL;

					DataForm *df;
					ValueRef *vr;
					EntryEntity *ee;
					gint pos;

					/* set the other QueryView */
					if (query_join_get_ant_view (qj) == qv)
						oqv = query_join_get_suc_view (qj);
					else
						oqv = query_join_get_ant_view (qj);
					
					/* set the other DbField */
					/* FIXME */
					/* if (QUERY_JOIN_PAIR (query_join_get_pairs (qj)->data)->ant_field ==  */
/* 					    G_OBJECT (qf)) */
/* 						ofield = DB_FIELD (QUERY_JOIN_PAIR (query_join_get_pairs (qj)->data)->suc_field); */
/* 					else */
/* 						ofield = DB_FIELD (QUERY_JOIN_PAIR (query_join_get_pairs (qj)->data)->ant_field); */

					/* set the other QueryField */
					oqf = query_get_field_by_dbfield (query_exec_get_query_ref (form->priv->qx), 
									  oqv, ofield);
					g_assert (oqv);
					g_assert (oqf);
					g_assert (ofield);
					
					if (! (df = data_form_find_from_query_view (form->priv->master, oqv))) {
						/* No data form found for oqv, build a new one */
						df = DATA_FORM (data_form_new_select (form->priv->qx, oqv, 
										      form->priv->numrow));
						df->priv->master = form->priv->master;
						form->priv->master->priv->children = g_slist_append (form->priv->master->priv->children,
												     df);
					}
				
					pos = query_get_pos_by_field (query_exec_get_query_ref (form->priv->qx), qf);
					/* EntryEntity */
					ee = g_new0 (EntryEntity, 1);
					ee->qf = qf;
					ee->numcol = pos;
					ee->entry = NULL;
					ee->form = df;
					ee->join_to = oqf;
					form->priv->entries = g_slist_append (form->priv->entries, ee);

					/* ValueRef */
					vr = data_form_find_value_ref (form, ofield);
					if (!vr) {
						vr = g_new0 (ValueRef, 1);
						vr->ref = ofield;
						vr->entry = ee;
							
						form->priv->value_refs = g_slist_append (form->priv->value_refs,
											 vr);
					}
					ee->value_ref = vr;
				}
				else { /* No QueryJoin for that QueryField */
					if (query_field_get_is_printed (qf) && 
					    !query_field_get_is_hidden (qf)) {
						DataEntry *de;
						DataHandler *dh;
						const GdaValue *value;
						DbField *field = NULL;
						EntryEntity *ee;
						ValueRef *vr;
						gint pos;
						ServerAccess *srv;
							
						pos = query_get_pos_by_field (query_exec_get_query_ref (form->priv->qx),
									      qf);
						if (form->priv->mode == DATA_FORM_EDIT_MODE)
							value = server_resultset_get_gdavalue (query_exec_get_data_set (form->priv->qx),
											       form->priv->numrow, pos);
						else
							value = NULL;

						srv = query_exec_get_query_ref (form->priv->qx)->conf->srv;
						field = query_field_field_get_db_field (qf);
						if (server_access_is_object_bound (srv, G_OBJECT (qf)))
							dh = server_access_get_object_handler (srv, 
											       G_OBJECT (qf));
						else
							dh = server_access_get_object_handler (srv, 
											       G_OBJECT (field));
						if (value) 
							de = DATA_ENTRY (data_handler_get_widget_from_value (dh, value));
						else {
							GdaValue *myvalue;
							myvalue = gda_value_new_null ();
							de = DATA_ENTRY (data_handler_get_widget_from_value (dh, myvalue));
							gda_value_free (myvalue);
						}
						data_entry_set_query_field (de, qf);
						
						if (form->priv->mode == DATA_FORM_INSERT_MODE) {
							if (field->default_value) 
								data_entry_set_is_default_set (de, TRUE);
							else {
								data_entry_set_is_default_set (de, FALSE);
								data_entry_set_is_default_allowed (de, FALSE);
							}
						}
						if (! field->null_allowed)
							data_entry_set_is_null_allowed (de, FALSE);

						data_entry_refresh_display (de);
						
						/* EntryEntity */
						ee = g_new0 (EntryEntity, 1);
						ee->qf = qf;
						ee->numcol = pos;
						ee->entry = de;
						ee->form = NULL;
						ee->join_to = NULL;
						form->priv->entries = g_slist_append (form->priv->entries, ee);

						/* ValueRef */
						vr = data_form_find_value_ref (form, field);
						if (!vr) {
							vr = g_new0 (ValueRef, 1);
							vr->ref = field;
							vr->entry = ee;
							pos = query_get_pos_by_field (query_exec_get_query_ref (form->priv->qx), qf);
							vr->entry = ee;
								
							form->priv->value_refs = g_slist_append (form->priv->value_refs,
												 vr);
						}
						ee->value_ref = vr;
					}
				}
			}
		}
		list = g_slist_next (list);
	}
}








/*
 *
 * INSERT MODE
 *
 */

static void data_form_insert_initialize (DataForm * form)
{
	GtkWidget *topwid = NULL;


	build_edit_insert_form_structure (form);
	
	if (0) {
		/* FIXME: update when glade forms are introduced */
		/* we have a custom glade form here */
		topwid = gtk_label_new ("Glade imported form to create");
		gtk_widget_show (topwid);
	}
	else {
		/* default builtin form */
		topwid = build_complete_noglade_form_widget (form);
		gtk_widget_show (topwid);
	}


	if (topwid) {
		GtkWidget *buttons;
		gtk_widget_show (topwid);
		gtk_box_pack_start_defaults (GTK_BOX (form), topwid);
		
		buttons = data_form_add_action_buttons (form, G_CALLBACK (data_form_user_action_cb), form);
		gtk_widget_show_all (buttons);
		gtk_box_pack_start (GTK_BOX (form), buttons, FALSE, FALSE, GNOME_PAD/2.);
	}
}









/*
 *
 * BROWSE MODE
 *
 */

static void data_form_browse_initialize (DataForm * form)
{

}







/*
 *
 * SELECT MODE
 *
 */

static void build_select_form_structure (DataForm * form);
static void data_form_select_initialize (DataForm * form)
{
	GtkWidget *topwid = NULL;

	if (!query_exec_get_data_set (form->priv->qx)) 
		topwid = gtk_label_new (_("No Data!"));
	else {
		build_select_form_structure (form);

		if (0) {
			/* we have a custom glade form here */
			topwid = gtk_label_new ("Glade imported form to create");
		}
		else {
			/* default builtin form */
			topwid = build_complete_noglade_form_widget (form);
			gtk_widget_show (topwid);
		}
	}

	if (topwid) {
		gtk_widget_show (topwid);
		gtk_box_pack_start_defaults (GTK_BOX (form), topwid);
	}
}

static void 
build_select_form_structure (DataForm * form)
{

}











/*
 *
 * FUNCTIONS FOR more than ONE MODES
 *
 */
static void data_entry_contents_modified_cb (DataEntry *de, DataForm *form);
static GtkWidget *
build_complete_noglade_form_widget (DataForm * form)
{
	GSList *list;
	guint i = 0;
	GtkWidget *topwid;
	
	topwid = gtk_table_new (g_slist_length (form->priv->entries), 2, FALSE);
	gtk_table_set_row_spacings (GTK_TABLE (topwid), 0);
	
	list = form->priv->entries;
	while (list) {
		GtkWidget *wid = NULL, *label;
		gchar *str;

		if (ENTRY_ENTITY (list->data)->entry) {
			wid = GTK_WIDGET (ENTRY_ENTITY (list->data)->entry);
			gtk_widget_show (wid);

			/* fetch the "contents_modified" signal */
			g_signal_connect (wid, "contents_modified",
					  G_CALLBACK (data_entry_contents_modified_cb), form);
		}
		if (ENTRY_ENTITY (list->data)->form) {
			wid = GTK_WIDGET (ENTRY_ENTITY (list->data)->form);
			gtk_widget_show (wid);
		}

		gtk_table_attach_defaults (GTK_TABLE (topwid), wid,
					   1, 2, i, i+1);
		
		if (query_field_get_name (ENTRY_ENTITY (list->data)->qf))
			str = g_strdup_printf ("%s:", query_field_get_name (ENTRY_ENTITY (list->data)->qf));
		else
			str = g_strdup_printf ("%s:", DB_FIELD (ENTRY_ENTITY (list->data)->value_ref->ref)->name);

		label = gtk_label_new (str);
		gtk_widget_show (label);
		g_free (str);
		
		gtk_table_attach (GTK_TABLE (topwid), label,
				  0, 1, i, i+1, 
				  GTK_FILL, 0, GNOME_PAD/2., 0);
		gtk_misc_set_alignment (GTK_MISC (label), 0, 0);
		
		list = g_slist_next (list);
		i++;
	}

	return topwid;
}




static void data_entry_refresh_actions_area (DataForm *form);
static void 
data_entry_contents_modified_cb (DataEntry *de, DataForm *form)
{
	data_entry_refresh_actions_area (form);
}


static gboolean data_form_is_on_first_entry (DataForm *form);
static gboolean data_form_is_on_last_entry (DataForm *form);
static gboolean data_form_is_diff_than_orig (DataForm *form);
static void 
data_entry_refresh_actions_area (DataForm *form)
{
	gchar *str;
	GtkWidget *wid, *collector;
	gboolean is_first, is_last, is_modified, to_commit;

	collector = form->priv->actions_area;
	if (!collector)
		return;

	if ((form->priv->mode == DATA_FORM_EDIT_MODE) &&
	    (server_resultset_get_nbtuples (query_exec_get_data_set (form->priv->qx)) == 0)) {
		is_first = TRUE;
		is_last = TRUE;
	}
	else {
		is_first = data_form_is_on_first_entry (form);
		is_last = data_form_is_on_last_entry (form);
	}

	is_modified = data_form_is_diff_than_orig (form);

	/* make sure there is no missing value in the form */
	to_commit = is_modified;
	if (to_commit) {
		GSList *list = form->priv->entries;

		while (list && to_commit) {
			if (ENTRY_ENTITY (list->data)->entry) {
				DataEntry *de = DATA_ENTRY (ENTRY_ENTITY (list->data)->entry);
				if (!data_entry_get_is_null_allowed (de) && 
				    data_entry_is_value_null (de) && 
				    !data_entry_is_value_default (de))
					to_commit = FALSE;
			}

			list = g_slist_next (list);
		}
	}

	if ((form->priv->mode == DATA_FORM_EDIT_MODE) &&
	    (server_resultset_get_nbtuples (query_exec_get_data_set (form->priv->qx)) == 0)) {
		to_commit = FALSE;
		is_modified = TRUE;
	}



	/* QUERY_ACTION_FIRST */
	str = g_strdup_printf ("w%d", QUERY_ACTION_FIRST);
	wid = g_object_get_data (G_OBJECT (collector), str);
	g_free (str);
	if (wid)
		gtk_widget_set_sensitive (wid, !is_first);

	/* QUERY_ACTION_PREV */
	str = g_strdup_printf ("w%d", QUERY_ACTION_PREV);
	wid = g_object_get_data (G_OBJECT (collector), str);
	g_free (str);
	if (wid)
		gtk_widget_set_sensitive (wid, !is_first);

	/* QUERY_ACTION_LAST */
	str = g_strdup_printf ("w%d", QUERY_ACTION_LAST);
	wid = g_object_get_data (G_OBJECT (collector), str);
	g_free (str);
	if (wid)
		gtk_widget_set_sensitive (wid, !is_last);

	/* QUERY_ACTION_NEXT */
	str = g_strdup_printf ("w%d", QUERY_ACTION_NEXT);
	wid = g_object_get_data (G_OBJECT (collector), str);
	g_free (str);
	if (wid)
		gtk_widget_set_sensitive (wid, !is_last);


	/* QUERY_ACTION_COMMIT */
	str = g_strdup_printf ("w%d", QUERY_ACTION_COMMIT);
	wid = g_object_get_data (G_OBJECT (collector), str);
	g_free (str);
	if (wid)
		gtk_widget_set_sensitive (wid, to_commit);

	/* QUERY_ACTION_DELETE */
	str = g_strdup_printf ("w%d", QUERY_ACTION_DELETE);
	wid = g_object_get_data (G_OBJECT (collector), str);
	g_free (str);
	if (wid)
		gtk_widget_set_sensitive (wid, !is_modified);
}

static gboolean
data_form_is_on_first_entry (DataForm *form)
{
	gboolean retval;
	g_return_val_if_fail (form && IS_DATA_FORM (form), FALSE);

	if (! query_exec_get_data_set (form->priv->qx) || 
	    (server_resultset_get_nbtuples (query_exec_get_data_set (form->priv->qx)) == 0))
		retval = TRUE;
	else {
		if (form->priv->numrow == 0)
			retval = TRUE;
		else
			retval = FALSE;
	}
	

	return retval;
}

static gboolean
data_form_is_on_last_entry  (DataForm *form)
{
	gboolean retval;
	g_return_val_if_fail (form && IS_DATA_FORM (form), FALSE);

	if (! query_exec_get_data_set (form->priv->qx) || 
	    (server_resultset_get_nbtuples (query_exec_get_data_set (form->priv->qx)) == 0))
		retval = TRUE;
	else {
		if (form->priv->numrow == server_resultset_get_nbtuples (query_exec_get_data_set (form->priv->qx))-1)
			retval = TRUE;
		else
			retval = FALSE;
	}

	return retval;
}

static gboolean 
data_form_is_diff_than_orig (DataForm *form)
{
	gboolean retval = FALSE;
	GSList *list;

	g_return_val_if_fail (form && IS_DATA_FORM (form), FALSE);

	list = form->priv->entries;
	while (list) {
		EntryEntity *ee;

		ee = ENTRY_ENTITY (list->data);

		if (ee->entry) {
			if (data_entry_is_value_modified (DATA_ENTRY (ee->entry)))
				retval = TRUE;
		}
		
		if (ee->form) { /* refresh the sub form */
			if (data_form_is_diff_than_orig (ee->form))
				retval = TRUE;
		}
		
		list = g_slist_next (list);
	}

	return retval;
}


void
data_form_refresh (DataForm *form)
{
	GSList *list;
	ServerResultset *rs;

	g_return_if_fail (form && IS_DATA_FORM (form));
	if ((form->priv->mode != DATA_FORM_EDIT_MODE) &&
	    (form->priv->mode != DATA_FORM_BROWSE_MODE))
		return;

	rs = query_exec_get_data_set (form->priv->qx);

	if (!rs || (rs && (server_resultset_get_nbtuples (rs) == 0))) {
		/* Show the "No Data" Page */
		gtk_notebook_set_current_page (GTK_NOTEBOOK (form->priv->top_notebook), 0);
		data_entry_refresh_actions_area (form);
	}
	else {
		/* Show the "actual Form" page */
		gtk_notebook_set_current_page (GTK_NOTEBOOK (form->priv->top_notebook), 1);

		if (form->priv->numrow > server_resultset_get_nbtuples (rs) - 1)
			form->priv->numrow = server_resultset_get_nbtuples (rs) - 1;

		list = form->priv->entries;
		while (list) {
			EntryEntity *ee;
			
			ee = ENTRY_ENTITY (list->data);
			
			if (ee->entry) {
				gint pos;
				const GdaValue *value;
				
				pos = query_get_pos_by_field (query_exec_get_query_ref (form->priv->qx), 
							      ee->qf);
				value = server_resultset_get_gdavalue (rs, form->priv->numrow, pos);
				data_entry_update_with_value (ee->entry, value, GDA_VALUE_TYPE_NULL, TRUE);
			}
			
			if (ee->form) { /* refresh the sub form */
				data_form_refresh (ee->form);
			}
		
			list = g_slist_next (list);
		}
	}
}

static GdaValue *
data_form_get_gdavalue (DataForm *df, QueryField *qf)
{
	/* FIXME */

	return NULL;
}

static DataForm *
data_form_find_from_query_view (DataForm *start, QueryView *qv)
{
	DataForm *retval = NULL;
	GSList *list;

	g_return_val_if_fail (start && IS_DATA_FORM (start), NULL);
	g_return_val_if_fail (qv && IS_QUERY_VIEW (qv), NULL);

	list = start->priv->children;
	while (list && !retval) {
		if (DATA_FORM (list->data)->priv->qv == qv)
			retval = DATA_FORM (list->data);
		list = g_slist_next (list);
	}

	return retval;
}

static ValueRef *
data_form_find_value_ref (DataForm *df, gpointer ref)
{
	ValueRef *vr = NULL;
	GSList *list;

	g_return_val_if_fail (df && IS_DATA_FORM (df), NULL);

	list = df->priv->value_refs;
	while (list && !vr) {
		if (VALUE_REF (list->data)->ref == ref)
			vr = VALUE_REF (list->data);
		list = g_slist_next (list);
	}

	return vr;
}


static void form_do_commit (DataForm *form);
static void form_do_delete (DataForm *form);
static void 
data_form_user_action_cb (GtkWidget *wid, DataForm * form)
{
	QueryEnvUserActions action;

	action = GPOINTER_TO_UINT (g_object_get_data (G_OBJECT (wid), "action_id"));
	switch (action) {

	case QUERY_ACTION_INSERT:
		g_signal_emit (G_OBJECT (form), data_form_signals [ACTION_INSERT], 0);
		break;
	case QUERY_ACTION_REFRESH:
		g_signal_emit (G_OBJECT (form), data_form_signals [ACTION_REFRESH], 0);
		break;
	case QUERY_ACTION_VIEWALL:
		g_signal_emit (G_OBJECT (form), data_form_signals [ACTION_VIEWALL], 0);
		break;
	case QUERY_ACTION_FIRST:
		form->priv->numrow = 0;
		data_form_refresh (form);
		break;
	case QUERY_ACTION_LAST:
		form->priv->numrow = server_resultset_get_nbtuples (query_exec_get_data_set (form->priv->qx))-1;
		data_form_refresh (form);
		break;
	case QUERY_ACTION_PREV:
		if (form->priv->numrow >= 1)
			form->priv->numrow--;
		data_form_refresh (form);
		break;
	case QUERY_ACTION_NEXT:
		if (form->priv->numrow+1 < server_resultset_get_nbtuples (query_exec_get_data_set (form->priv->qx)))
			form->priv->numrow++;
		data_form_refresh (form);
		break;
	case QUERY_ACTION_COMMIT:
		form_do_commit (form);
		break;
	case QUERY_ACTION_DELETE:
		form_do_delete (form);
		break;
	case QUERY_ACTION_EDIT:
		g_print ("To be implemented in %s line %d\n", __FILE__, __LINE__);
		break;
	default :
		break;
	}
}


static gchar * make_update_sql_query (DataForm *form, GSList *param_values);
static gchar * make_insert_sql_query (DataForm *form, GSList *param_values);
static void 
form_do_commit (DataForm *form)
{
	GSList *list, *values = NULL;
	gchar *query;
	ServerResultset *rs;

	/* build the new values list */
	list = form->priv->entries;
	while (list) {
		QueryParamValue *value;
		
		if (ENTRY_ENTITY (list->data)->entry) {
			DataEntry *de = ENTRY_ENTITY (list->data)->entry;
			
			value = g_new0 (QueryParamValue, 1);
			value->field = ENTRY_ENTITY (list->data)->qf;
			value->value = data_entry_get_gdavalue (de);
			value->dh = data_entry_get_data_handler (de);

			/* QF type test */
			if (query_field_get_ftype (value->field) != QUERY_FIELD_FIELD)
				g_warning ("EntryEntity %p does not reference a DbField\n", list->data);
			
			if (data_entry_is_value_default (de)) {
				/* fetch the default GdaValue in the DbField */
				DbField *field = query_field_field_get_db_field (value->field);
				if (field->default_value) 
					value->value = gda_value_copy (field->default_value);
			}
			else {
				/* value from the DataEntry widget (if the value is NULL, 
				   then the function will return a GdaValue for NULL) */
				value->value = data_entry_get_gdavalue (de);
			}
				
		}
		else {
			/* FIXME : data handler: get the default one for the right data type */
			DataForm *subform = ENTRY_ENTITY (list->data)->form;
			value = g_new0 (QueryParamValue, 1);
			value->field = ENTRY_ENTITY (list->data)->qf;
			value->value = data_form_get_gdavalue (subform, value->field);
		}
		values = g_slist_append (values, value);

		list = g_slist_next (list);
	}

	/* build the SQL statement */
	switch  (form->priv->mode) {
	case DATA_FORM_EDIT_MODE:
		query = make_update_sql_query (form, values);
		break;
	case DATA_FORM_INSERT_MODE:
		query = make_insert_sql_query (form, values);
		break;
	default:
		query = NULL;
		g_warning ("form_do_commit () called on the wrong mode: %d\n", form->priv->mode);
	}

	/* free all the values */
	list = values;
	while (list) {
		gda_value_free (QUERY_PARAM_VALUE (list->data)->value);
		g_free (list->data);
		list = g_slist_next (list);
	}
	g_slist_free (values);

	/* Query execution */
	rs = server_access_do_query (query_view_get_query (form->priv->qv)->conf->srv, query, 
				     SERVER_ACCESS_QUERY_SQL);
	if (rs) {
		g_signal_emit (G_OBJECT (form), data_form_signals [ACTION_REFRESH], 0);
		
		if (form->priv->mode == DATA_FORM_INSERT_MODE) {
			/* close the form */
			gtk_widget_destroy (GTK_WIDGET (form));
		}
		
		g_object_unref (G_OBJECT (rs));
	}
	
	g_free (query);
}

static gchar * make_where_cond_sql_part (DataForm *form);
static gchar *
make_update_sql_query (DataForm *form, GSList *param_values)
{
	GString *string;
	gchar *str;
	DbTable *modif_table;
	GSList *list;
	gboolean first_field;

	g_return_val_if_fail (form->priv->mode == DATA_FORM_EDIT_MODE, NULL);
	string = g_string_new ("UPDATE ");

	modif_table = query_exec_get_modif_table (form->priv->qx);
	g_string_append_printf (string, "%s SET ", modif_table->name);

	/* New values */
	first_field = TRUE;
	list = param_values;
	while (list) {
		QueryParamValue *value;
		DbField *field;
		
		value = QUERY_PARAM_VALUE (list->data);
		field = query_field_field_get_db_field (value->field);

		if (first_field)
			first_field = FALSE;
		else
			g_string_append (string, ", ");

		if (gda_value_is_null (value->value))
			g_string_append_printf (string, "%s = NULL", field->name);
		else {
			gchar *str;
			if (value->dh)
				str = data_handler_get_sql_from_value (value->dh, value->value);
			else
				str = gda_value_stringify (value->value);
			g_string_append_printf (string, "%s = %s", field->name, str);
			g_free (str);
		}

		list = g_slist_next (list);
	}

	/* WHERE conditions */
	str = make_where_cond_sql_part (form);
	g_string_append (string, " ");
	g_string_append (string, str);
	g_free (str);

	str = string->str;
	g_string_free (string, FALSE);
	return str;
}

static gchar *
make_where_cond_sql_part (DataForm *form)
{
	gchar *str;
	GString *string;
	gboolean first_field = TRUE;
	GSList *list;
	DbTable *modif_table;


	modif_table = query_exec_get_modif_table (form->priv->qx);
	string = g_string_new ("WHERE ");

	/* WHERE conditions from the currently selected row in the recordset */
	list = query_exec_get_query_ref (form->priv->qx)->fields;
	while (list) {
		QueryField *qf;
		qf = QUERY_FIELD (list->data);

		if (query_field_get_ftype (qf) == QUERY_FIELD_FIELD) {
			DbField *field;
			DbTable *table;
			gint pos;
			const GdaValue *value;

			field = query_field_field_get_db_field (qf);
			table = database_find_table_from_field (query_view_get_query (form->priv->qv)->conf->db, field);
			if (table == modif_table) { /* WHERE condition only on the modified table */
				if (first_field)
					first_field = FALSE;
				else
					g_string_append (string, "AND ");
				
				g_string_append_printf (string, "%s ", field->name);
				pos = query_get_pos_by_field (query_exec_get_query_ref (form->priv->qx), qf);
				value = server_resultset_get_gdavalue (query_exec_get_data_set (form->priv->qx),
								       form->priv->numrow, pos);
				if (gda_value_is_null (value))
					g_string_append (string, "IS NULL ");
				else {
					gchar *str;
					DataHandler *dh;
					
					dh = server_access_get_object_handler (query_view_get_query (form->priv->qv)->conf->srv,
									       G_OBJECT (field->type));
					if (dh)
						str = data_handler_get_sql_from_value (dh, value);
					else
						str = gda_value_stringify (value);
					g_string_append_printf (string, "= %s ", str);
					g_free (str);
				}
			}
		}

		list = g_slist_next (list);
	}

	/* WHERE conditions coming from the original Query */
	/* FIXME: add the WHERE conditions coming from the original Query:
	   keep only the QueryCond related to the modified table? */

	str = string->str;
	g_string_free (string, FALSE);
	return str;
}

static gchar *
make_insert_sql_query (DataForm *form, GSList *param_values)
{
	GString *string, *stringvals;
	gchar *str;
	DbTable *modif_table;
	GSList *list;
	gboolean first_field;

	g_return_val_if_fail (form->priv->mode == DATA_FORM_INSERT_MODE, NULL);
	string = g_string_new ("INSERT INTO ");
	stringvals = g_string_new ("VALUES (");

	modif_table = query_exec_get_modif_table (form->priv->qx);
	g_string_append_printf (string, "%s (", modif_table->name);

	/* Parsing the new values to create the insert query */
	first_field = TRUE;
	list = param_values;
	while (list) {
		QueryParamValue *value;
		DbField *field;
		
		value = QUERY_PARAM_VALUE (list->data);
		field = query_field_field_get_db_field (value->field);

		if (first_field)
			first_field = FALSE;
		else {
			g_string_append (string, ", ");
			g_string_append (stringvals, ", ");
		}

		if (gda_value_is_null (value->value)) {
			g_string_append_printf (string, "%s", field->name);
			g_string_append (stringvals, "NULL");
		}
		else {
			gchar *str;

			g_string_append_printf (string, "%s", field->name);
			if (value->dh)
				str = data_handler_get_sql_from_value (value->dh, value->value);
			else
				str = gda_value_stringify (value->value);
			g_string_append_printf (stringvals, "%s", str);
			g_free (str);
		}

		list = g_slist_next (list);
	}
	g_string_append (string, ") ");
	g_string_append (stringvals, ")");

	g_string_append (string, stringvals->str);

	str = string->str;
	g_string_free (string, FALSE);
	g_string_free (stringvals, TRUE);
	return str;
}


static gchar * make_delete_sql_query (DataForm *form);

/* delete the current selected entry */
static void 
form_do_delete (DataForm *form)
{
	gchar *query;
	ServerResultset *rs;

	/* build the SQL statement */
	switch  (form->priv->mode) {
	case DATA_FORM_EDIT_MODE:
		query = make_delete_sql_query (form);
		break;
	default:
		query = NULL;
		g_warning ("form_do_delete () called on the wrong mode: %d\n", form->priv->mode);
	}


	rs = server_access_do_query (query_view_get_query (form->priv->qv)->conf->srv, query, SERVER_ACCESS_QUERY_SQL);
	if (rs) {
		g_signal_emit (G_OBJECT (form), data_form_signals [ACTION_REFRESH], 0);
		g_object_unref (G_OBJECT (rs));
	}

	g_free (query);
}

static gchar *
make_delete_sql_query (DataForm *form)
{
	gchar *str;
	GString *string;
	DbTable *modif_table;

	string = g_string_new ("DELETE FROM ");
	modif_table = query_exec_get_modif_table (form->priv->qx);
	g_string_append_printf (string, "%s ", modif_table->name);
	
	str = make_where_cond_sql_part (form);
	g_string_append (string, str);
	g_free (str);

	str = string->str;
	g_string_free (string, FALSE);
	return str;
}









/*
 * Actions buttons
 */
static void add_stock_pixmap (GtkWidget * button_box, gchar * action,
			      gchar * tip, QueryEnvUserActions action_val,
			      GCallback cb, gpointer data,
			      GObject *collector);

static GtkWidget *
data_form_add_action_buttons (DataForm *form, GCallback cb, gpointer data)
{
	GtkWidget *vbox, *bbox;
	gint actions = form->priv->actions;

	vbox = gtk_vbox_new (FALSE, GNOME_PAD / 4.);
	/* first row */
	bbox = gtk_hbutton_box_new ();
	gtk_box_pack_start (GTK_BOX (vbox), bbox, FALSE, FALSE, 0);
	gtk_button_box_set_layout (GTK_BUTTON_BOX (bbox), GTK_BUTTONBOX_START);
	gtk_button_box_set_spacing (GTK_BUTTON_BOX (bbox), GNOME_PAD / 4.);

	if (actions & QUERY_ACTION_FIRST)
		add_stock_pixmap (bbox, GTK_STOCK_GOTO_FIRST,
				  _("Move to the first record"),
				  QUERY_ACTION_FIRST, cb, data, G_OBJECT (vbox));

	if (actions & QUERY_ACTION_PREV)
		add_stock_pixmap (bbox, GTK_STOCK_GO_BACK,
				  _("Move to the previous record"),
				  QUERY_ACTION_PREV, cb, data, G_OBJECT (vbox));

	if (actions & QUERY_ACTION_NEXT)
		add_stock_pixmap (bbox, GTK_STOCK_GO_FORWARD,
				  _("Move to the next record"),
				  QUERY_ACTION_NEXT, cb, data, G_OBJECT (vbox));

	if (actions & QUERY_ACTION_LAST)
		add_stock_pixmap (bbox, GTK_STOCK_GOTO_LAST,
				  _("Move to the last record"),
				  QUERY_ACTION_LAST, cb, data, G_OBJECT (vbox));


	/* second row */
	bbox = gtk_hbutton_box_new ();
	gtk_box_pack_start (GTK_BOX (vbox), bbox, FALSE, FALSE, 0);
	gtk_button_box_set_layout (GTK_BUTTON_BOX (bbox), GTK_BUTTONBOX_START);
	gtk_button_box_set_spacing (GTK_BUTTON_BOX (bbox), GNOME_PAD / 4.);

	if (actions & QUERY_ACTION_EDIT)
		add_stock_pixmap (bbox, GTK_STOCK_INDEX,
				  _("Edit data"),
				  QUERY_ACTION_EDIT, cb, data, G_OBJECT (vbox));

	if (actions & QUERY_ACTION_VIEWALL)
		add_stock_pixmap (bbox, GTK_STOCK_INDEX,
				  _("View all as grid"),
				  QUERY_ACTION_VIEWALL, cb, data, G_OBJECT (vbox));

	if (actions & QUERY_ACTION_INSERT)
		add_stock_pixmap (bbox, GTK_STOCK_NEW,
				  _("Insert new data"),
				  QUERY_ACTION_INSERT, cb, data, G_OBJECT (vbox));

	if (actions & QUERY_ACTION_COMMIT)
		add_stock_pixmap (bbox, GTK_STOCK_SAVE,
				  _("Save changes"),
				  QUERY_ACTION_COMMIT, cb, data, G_OBJECT (vbox));

	if (actions & QUERY_ACTION_DELETE)
		add_stock_pixmap (bbox, GTK_STOCK_REMOVE,
				  _("Delete selected data"),
				  QUERY_ACTION_DELETE, cb, data, G_OBJECT (vbox));

	if (actions & QUERY_ACTION_REFRESH)
		add_stock_pixmap (bbox, GTK_STOCK_REFRESH,
				  _("Refresh display"),
				  QUERY_ACTION_REFRESH, cb, data, G_OBJECT (vbox));


	/* Necessary settings to be set */
	form->priv->actions_area = vbox;
	data_entry_refresh_actions_area (form);

	return vbox;
}

static void
add_stock_pixmap (GtkWidget * button_box, gchar * action,
		  gchar * tip, QueryEnvUserActions action_val,
		  GCallback cb, gpointer data,
		  GObject *collector)
{
	GtkWidget *button, *pix;
	static GtkTooltips *tips = NULL;
	gchar *str;

	button = gtk_button_new ();
	pix = gtk_image_new_from_stock (action, GTK_ICON_SIZE_MENU);
	gtk_container_add (GTK_CONTAINER (button), pix);
	gtk_container_add (GTK_CONTAINER (button_box), button);

	if (!tips)
		tips = gtk_tooltips_new ();
	gtk_tooltips_set_tip (GTK_TOOLTIPS (tips), button, tip, NULL);
	gtk_tooltips_enable (GTK_TOOLTIPS (tips));

	str = g_strdup_printf ("w%d", action_val);
	g_object_set_data (G_OBJECT (collector), str, button);
	g_free (str);

	g_object_set_data (G_OBJECT (button), "action_id", GUINT_TO_POINTER (action_val));
	g_object_set_data (G_OBJECT (button), "collector", collector);
	
	g_signal_connect (G_OBJECT (button), "clicked", cb, data);
}
